
#ifndef CHAOS_H
#define CHAOS_H

#include <cstdlib>
#include <iostream>

#include "numtypes.h"
#include "WELL512a.h"
#include "primes.h"

struct Chaos {
  // real arrays only!
  
  // initial data
  real uniform [N_UNIFORM];
  real normal  [N_NORMAL];
  
  // topography
  real t_uniform [T_UNIFORM];
  real t_normal  [T_NORMAL];
  
  // coefficient k
  real k_uniform [K_UNIFORM];
  real k_normal  [K_NORMAL];
};

// support for deterministic runs when sizeof(Chaos) = 0
static const int CHAOS_SIZE = sizeof(Chaos) / sizeof(real) != 0 ? sizeof(Chaos) / sizeof(real) : 1;

extern Chaos chaos;
extern Chaos *iid;
extern long_int iid_size;

static real NORMAL_buffer;
static bool NORMAL_buffer_empty = true;

static void chaos_resize (long_int M) {
  iid_size = M;
  iid = (Chaos*) realloc (iid, iid_size*CHAOS_SIZE*sizeof(real));
  if (iid == NULL) {
    std::cout << " :: ERROR: could not realloc iid in chaos_resize(). Requested size was: " << iid_size << std::endl;
    throw 1;
  }
}

// generates uniform random variable on [0, 1]
static real draw_UNIFORM () {
  return WELLRNG512a();
}

// generates standard normal random variable using rejection method and Box-Muller transformation
static real draw_NORMAL () {
  
  if (!NORMAL_buffer_empty) {
    NORMAL_buffer_empty = true;
    return NORMAL_buffer;
  }
  
  else {
    real U1, U2, r, w, N1, N2;
    
    // rejection method to obtain (U1,U2) - uniform random variable on unit circle
    do {
      // draw two uniform random variable on [-1, 1]
      U1 = 2*draw_UNIFORM() - 1;
      U2 = 2*draw_UNIFORM() - 1;
      // compute radius
      r = sqr(U1) + sqr(U2);
    } while (r > 1.0);
    
    // Box-Muller transformation
    w  = sqrt(-2*log(r)/r);
    N1 = U1 * w;
    N2 = U2 * w;
    
    NORMAL_buffer = N2;
    NORMAL_buffer_empty = false;
    
    return N1;
  }
}

static Chaos generate_chaos () {
  Chaos chaos;
  
  // initial data
  for (int i=0; i<N_UNIFORM; i++)
    chaos.uniform[i] = draw_UNIFORM();
  for (int i=0; i<N_NORMAL; i++)
    chaos.normal[i] = draw_NORMAL();
  
  // topography
  for (int i=0; i<T_UNIFORM; i++)
    chaos.t_uniform[i] = draw_UNIFORM();
  for (int i=0; i<T_NORMAL; i++)
    chaos.t_normal[i] = draw_NORMAL();
  
  // coefficient k
  for (int i=0; i<K_UNIFORM; i++)
    chaos.k_uniform[i] = draw_UNIFORM();
  for (int i=0; i<K_NORMAL; i++)
    chaos.k_normal[i] = draw_NORMAL();
  
  return chaos;
}

static void chaos_init (int offset=0) {
  // offsetting samples
  for (long_int i=0; i<offset; i++)
    generate_chaos ();
  
  // generating offsetted samples
  for (long_int i=0; i<iid_size; i++)
    iid[i] = generate_chaos ();
}

static void chaos_buffer_init (int seed) {
  unsigned int buffer[getR()];
  
  // init buffer using Linear Congruential Generator
  
  const long_int a = 1103515245;
  const long_int c = 12345;
  
  unsigned int x = seed;
  
  for (int i=0; i<getR(); i++) {
    x = a*x + c;
    buffer[i] = x;
  }
  
  InitWELLRNG512a (buffer);
}

#endif
