
#include "main.h"
#include "solver.h"
#include "ec_util.h"

///::title:: ES solver (Energy Stable) | for Shallow Water
///::consts:: ENTROPIC_RECONSTRUCTION=1
///::depends:: ec_util

void solver_calc_flux(ConservedVars &F, vector &us, vector &Bs,
                      const AllVariables &lV, const AllVariables &rV,
                      int i, int j, int k, int d)
{
  // construction of jump in EntropyVars
  
  real V [3];
  
  //V[0] = jump ( GRAVITY * (lV.U.h + lV.XU.b) - ( sqr(lV.XU.u[X]) + sqr(lV.XU.u[Y]) ) / 2.0, GRAVITY * (rV.U.h + rV.XU.b) - ( sqr(rV.XU.u[X]) + sqr(rV.XU.u[Y]) ) / 2.0 );
  V[0] = jump ( lV.U.h    , rV.U.h     );
  V[1] = jump ( lV.XU.u[X], rV.XU.u[X] );
  V[2] = jump ( lV.XU.u[Y], rV.XU.u[Y] );
  
  // for here on, ConservedVars are needed
  
  ConservedVars lCV, rCV;
  XtraVars      lXV, rXV;
  
  // for EC flux, we get ConservedVars from EntropyVars U, Xtra
  
  lCV = U[i-DX[d]][j-DY[d]][k];
  rCV = U[i][j][k];
  lXV = Xtra[i-DX[d]][j-DY[d]][k];
  rXV = Xtra[i][j][k];
  
  convert_to_ConservedVars (lCV, lXV);
  convert_to_ConservedVars (rCV, rXV);
  
  // === EC flux
  
  //ec_flux (F, lV, rV, i, j, k, d);
  ec_flux (F, lCV, rCV, lXV, rXV, d);
  
  // for numerical diffusion operator, we get ConservedVars from EntropyVars lV, rV
  
  lCV = lV.U;
  rCV = rV.U;
  lXV = lV.XU;
  rXV = rV.XU;
  
  convert_to_ConservedVars (lCV, lXV);
  convert_to_ConservedVars (rCV, rXV);
  
  // === numerical diffusion
  
  // construction of matrix R
  
  real R [3][3] = { {1, 0, 1}, {0, 0, 0}, {0, 0, 0} };
  
  real sqrt_gh = sqrt ( GRAVITY * avg(lCV.h, rCV.h) );
  
  // 1st column of R
  R[1+d][0] = avg(lXV.u[d], rXV.u[d]) - sqrt_gh;
  R[2-d][0] = avg(lXV.u[1-d], rXV.u[1-d]);
  
  // 2nd column of R
  if (d == X)
    R[2][1] = sqrt_gh;
  else // (d == Y)
    R[1][1] = -sqrt_gh;
  
  // 3rd column of R
  R[1+d][2] = avg(lXV.u[d], rXV.u[d]) + sqrt_gh;
  R[2-d][2] = avg(lXV.u[1-d], rXV.u[1-d]);
  
  real scaling = sqrt(2*GRAVITY);
  for (int a=0; a<3; a++)
    for (int b=0; b<3; b++)
      R[a][b] /= scaling;
  
  // construction of diagonal D for Roe type diffusion
  
  real D [3];
  
  D[0] = fabs ( avg(lXV.u[d], rXV.u[d]) - sqrt_gh );
  D[1] = fabs ( avg(lXV.u[d], rXV.u[d])           );
  D[2] = fabs ( avg(lXV.u[d], rXV.u[d]) + sqrt_gh );
  
  // computing diffusion matrix RDR'
  
  // DR'
  real DR  [3][3];
  
  for (int a=0; a<3; a++)
    for (int b=0; b<3; b++)
      DR[a][b] = D[a] * R[b][a];
  
  // R(DR')
  real RDR [3][3] = {0.0};
  
  for (int a=0; a<3; a++)
    for (int b=0; b<3; b++)
      for (int c=0; c<3; c++)
        RDR[a][b] += R[a][c] * DR[c][b];
  
  // computing diffusion term
  
  real RDRV [3] = {0.0};
  
  for (int a=0; a<3; a++)
    for (int b=0; b<3; b++)
      RDRV[a] += RDR[a][b] * V[b];
  
  // adding diffusion to EC flux
  
  F.h     -= 0.5 * RDRV[0];
  F.hu[X] -= 0.5 * RDRV[1];
  F.hu[Y] -= 0.5 * RDRV[2];
}
