
/* MHD (Magneto-HydroDynamics equations)
 * Structures and operator overloading for ConservedVars, PrimitiveVars and XtraVars
 */

#ifndef STRUCTURES_MHD_H
#define STRUCTURES_MHD_H

#include "numtypes.h"

// === ConservedVars

struct ConservedVars {
  real   rho;
  vector m;
  real   E;
  vector B;
  
  real norm() const
  {
    return fabs(rho)+m.norm()+fabs(E)+B.norm();
  }
  
  // defined in "structures.cpp"
  
  void operator += (const ConservedVars &a);
  void operator -= (const ConservedVars &a);
  void operator *= (const ConservedVars &a);
  void operator /= (const ConservedVars &a);
  void operator /= (const real r);
  void operator *= (const real r);
  void operator =  (const real r);
  #if COMPUTE_KURTOSIS
  void operator -= (const real r);
  #endif
};

// === XtraVars

struct XtraVars {
  real   p;
  vector u;
  real   a;
  vector cf;
  
  // defined in "structures.cpp"
  
  void operator += (const XtraVars &a);
  void operator -= (const XtraVars &a);
  void operator *= (const XtraVars &a);
  void operator /= (const XtraVars &a);
  void operator /= (const real r);
  void operator *= (const real r);
  void operator =  (const real r);
  #if COMPUTE_KURTOSIS
  void operator -= (const real r);
  #endif
};

// === PrimitiveVars

struct PrimitiveVars {
  real   rho;
  real   p;
  vector u;
  vector B;
  
  void toConservedVars(ConservedVars &v) const {
    v.rho = rho;
    v.m = rho*u;
    v.E = p/(GAMMA-1) + 0.5*rho*u.sqr() + 0.5*B.sqr();
    v.B = B;
  }
  
  PrimitiveVars() {}
  PrimitiveVars(const ConservedVars &CV, const XtraVars &XV);
};

// picks PrimitiveVars from ConservedVars (U) and XtraVars (Xtra)
inline PrimitiveVars::PrimitiveVars(const ConservedVars &CV, const XtraVars &XV)
{
  rho = CV.rho;
  p   = XV.p;
  u   = XV.u;
  B   = CV.B;
}

#endif
